%
%     Implementation by Piotr Hoownia <holownia@student.agh.edu.pl>
%
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     PLNXT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of PLNXT.
%
%     PLNXT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     PLNXT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with PLNXT.  If not, see <http://www.gnu.org/licenses/>.
%

:- module(nxt_actions_serial,[
				nxt_actions_connection_open/0,
				nxt_actions_connection_close/0,
				nxt_actions_motor/3,
				nxt_actions_motor_sync/5,
				nxt_actions_motor_is_moving/1,
				nxt_actions_motor_stop/1,
			      	nxt_actions_motor_stop/2,
			      	nxt_actions_motor_get_speed/2,
			      	nxt_actions_motor_get_rotations/2,
			      	nxt_actions_motor_reset_position/2,
				nxt_actions_touch_sensor/2,
				nxt_actions_sound_sensor/2,
				nxt_actions_light_sensor/2,
				nxt_actions_light_sensor_LED/2,
				nxt_actions_ultrasonic_sensor/2,
				nxt_actions_voltage/1,
				nxt_actions_voltage_millivolt/1,
				nxt_actions_start_program/1,
				nxt_actions_stop_program/0,
				nxt_actions_play_sound_file/2,
				nxt_actions_stop_sound_playback/0,
				nxt_actions_play_tone/2,
				nxt_actions_get_brick_name/1,
				nxt_actions_set_brick_name/1
				]).

/** <module> NXT Mindstroms - Actions, serial.

Motors: 'A', 'B', 'C'.
Sensors ports: 'S1', 'S2', 'S3', 'S4'.

@author Piotr Hoownia
@license GNU General Public License
*/

:- dynamic nxt_actions_streams/2.

:- dynamic nxt_actions_input_set/2,
	nxt_actions_input_set/3.


%%	nxt_actions_connection_open
%
%	Opens connection.

nxt_actions_connection_open :-
	nxt_serial_device(write,Write_device),
	nxt_serial_device(read,Read_device),
	%concat('stty -F ',Device,A),
	%concat(A,' -echo -icanon -iexten -isig -icrnl -inpck -istrip -ixon -cstopb cs8 parenb parodd -opost ispeed 115200 ospeed 115200',Cmd),
	%shell(Cmd),
	open(Write_device,write,W,[buffer(false)]),
	open(Read_device,read,R,[buffer(true)]),
	retractall(nxt_actions_streams(write,_)),
	retractall(nxt_actions_streams(read,_)),
	assert(nxt_actions_streams(write,W)),	
	assert(nxt_actions_streams(read,R)).


%%	nxt_actions_connection_close
%
%	Closes connection.

nxt_actions_connection_close :-
	nxt_actions_streams(write,W),
	nxt_actions_streams(read,R),
	retractall(nxt_actions_streams(write,_)),
	retractall(nxt_actions_streams(read,_)),
	close(W),
	close(R).


%%	nxt_actions_motor(+Motor,+Speed,+Angle).
%
%	Rotates motor at specified speed.
%	Forward if Speed is positive, backward if Speed is negative.
%	Motor will stop, when specified revolution (Angle in degrees) is reached.
%	Motor will never stop, if Angle equals 0.

nxt_actions_motor(Motor,Speed,Angle) :-
	nxt_actions_motor_move(Motor,Speed,[motoron,regulated],speed,0,running,Angle).


%%	nxt_actions_motor_sync(+Motor1,+Motor2,+Speed,+Turn_ratio,+Angle).
%
%	Rotates Motor1 at specified speed.
%	Motor2 is synchronized with Motor1 with specified Turn_ratio.
%	Motors will stop, when specified revolution (Angle in degrees) is reached.
%	Motors will never stop, if Angle equals 0.
%FIXME Opisz, o co chodzi z Turn_ratio!!!

%Przyklad: nxt_actions_motor_sync('C','B',450,5,2000).
nxt_actions_motor_sync(Motor1,Motor2,Speed,Turn_ratio,Angle) :-
	nxt_actions_motor_stop(Motor1),
	nxt_actions_motor_stop(Motor2),
	nxt_actions_motor_reset_position(Motor1,last),
	nxt_actions_motor_reset_position(Motor2,last),
	nxt_actions_motor_move(Motor1,Speed,[motoron,brake,regulated],sync,Turn_ratio,running,Angle),
	nxt_actions_motor_move(Motor2,Speed,[motoron,brake,regulated],sync,Turn_ratio,running,Angle).


%%	nxt_actions_motor_is_moving(+Motor).
%
%	Returns true if specified motor is moving. Otherwise fails.

nxt_actions_motor_is_moving(Motor) :-
	nxt_actions_motor_get_speed(Motor,Speed),
	Speed \= 0.


%%	nxt_actions_motor_stop(+Motor).
%
%	Stops motor.

nxt_actions_motor_stop(Motor) :-
	nxt_actions_motor_stop(Motor,true).


%%	nxt_actions_motor_stop(+Motor,+Brake).
%
%	Stops motor.

nxt_actions_motor_stop(Motor,true) :-
	nxt_actions_motor_move(Motor,0,[motoron,brake,regulated],speed,0,running,0).

nxt_actions_motor_stop(Motor,false) :-
	nxt_actions_motor_move(Motor,0,0,idle,0,idle,0).


%%	nxt_actions_motor_get_speed(+Motor,-Speed).
%
%	Gets speed of the motor.

nxt_actions_motor_get_speed(Motor,Speed) :-
	nxt_actions_connection_type(bt),
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(3),
	nxt_actions_motor_get_output_cmd(Motor),
	nxt_actions_pop_byte(2),
	nxt_actions_motor_get_output_response(Speed,_),
	mutex_unlock(nxt_serial_device).

nxt_actions_motor_get_speed(Motor,Speed) :-
	nxt_actions_connection_type(usb),
	mutex_lock(nxt_serial_device),
	nxt_actions_motor_get_output_cmd(Motor),
	nxt_actions_motor_get_output_response(Speed,_),
	mutex_unlock(nxt_serial_device).


%%	nxt_actions_motor_get_rotations(+Motor,-Rot_count).
%
%	Gets rotation count of the motor.

nxt_actions_motor_get_rotations(Motor,Rot_count) :-
	nxt_actions_connection_type(bt),
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(3),
	nxt_actions_motor_get_output_cmd(Motor),
	nxt_actions_pop_byte(2),
	nxt_actions_motor_get_output_response(_,Rot_count),
	mutex_unlock(nxt_serial_device).

nxt_actions_motor_get_rotations(Motor,Rot_count) :-
	nxt_actions_connection_type(usb),
	mutex_lock(nxt_serial_device),
	nxt_actions_motor_get_output_cmd(Motor),
	nxt_actions_motor_get_output_response(_,Rot_count),
	mutex_unlock(nxt_serial_device).

nxt_actions_motor_get_output_cmd(Motor) :-
	nxt_actions_output_port(Motor,Output_port),
	nxt_actions_send_bytes(['\x00','\x06',Output_port]).

%nxt_actions_motor_get_output_response(S,R) :-
%	nxt_actions_get_bytes_debug(25).

nxt_actions_motor_get_output_response(Speed,Rot_count) :-
	nxt_actions_pop_byte(2),
	nxt_actions_check_status_byte, !,
	nxt_actions_pop_byte,
	nxt_actions_get_sbyte(Possible_power),
	nxt_actions_pop_byte(3),
	nxt_actions_motor_get_output_response_check_run_state(Possible_power,Power),
	Speed is Power*9,
	nxt_actions_pop_byte(12),
	nxt_actions_get_slong(Rot_count).

% Before starting the motor an error may occure (status byte ='\xC0').
nxt_actions_motor_get_output_response(0,0) :-
	nxt_actions_pop_byte(22).

% Motor can be in state: it is not moving, but power differs from 0. It can be recognized by checking run state.
nxt_actions_motor_get_output_response_check_run_state(_,Real_power) :-
	nxt_actions_get_byte(Run_state),
	Run_state = 0, !,
	Real_power = 0.

nxt_actions_motor_get_output_response_check_run_state(Real_power,Real_power).
	

%%	nxt_actions_motor_reset_position(+Motor,+Relative).
%
%	Resets motor position.
%	Relative is one of:
%
%		* last
%		Resets position relative to last movement.
%
%		* absolute
%		Resets absolute position.

nxt_actions_motor_reset_position(Motor,Relative) :-
	nxt_actions_connection_type(bt),
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(4),
	nxt_actions_motor_reset_position_cmd(Motor,Relative),
	mutex_unlock(nxt_serial_device).

nxt_actions_motor_reset_position(Motor,Relative) :-
	nxt_actions_connection_type(usb),
	mutex_lock(nxt_serial_device),
	nxt_actions_motor_reset_position_cmd(Motor,Relative),
	mutex_unlock(nxt_serial_device).

nxt_actions_motor_reset_position_cmd(Motor,absolute) :-
	nxt_actions_output_port(Motor,Output_port),
	nxt_actions_send_bytes(['\x80','\x0A',Output_port,0]).

nxt_actions_motor_reset_position_cmd(Motor,last) :-
	nxt_actions_output_port(Motor,Output_port),
	nxt_actions_send_bytes(['\x80','\x0A',Output_port,1]).


%%	nxt_actions_touch_sensor(+Port,-Value).
%
%	Gets touch sensor reading.
%	Returns 1 if pressed, 0 otherwise.

nxt_actions_touch_sensor(Port,Value) :-
	not(nxt_actions_input_set(touch,Port)), !,
	retractall(nxt_actions_input_set(touch,_)),
	nxt_actions_input_setmode(Port,switch,boolean),
	assert(nxt_actions_input_set(touch,Port)),
	nxt_actions_touch_sensor(Port,Value).

nxt_actions_touch_sensor(Port,Value) :-
	nxt_actions_input_get(Port,Value), !.


%%	nxt_actions_sound_sensor(+Port,-Value).
%
%	Gets sound sensor reading.

nxt_actions_sound_sensor(Port,Value) :-
	not(nxt_actions_input_set(sound,Port)), !,
	retractall(nxt_actions_input_set(sound,_)),
	nxt_actions_input_setmode(Port,sounddb,pctfullscale),
	assert(nxt_actions_input_set(sound,Port)),
	nxt_actions_sound_sensor(Port,Value).

nxt_actions_sound_sensor(Port,Value) :-
	nxt_actions_input_get(Port,Value), !.


%%	nxt_actions_light_sensor(+Port,-Value,+Setting).
%
%	Gets light sensor reading.
%	The LED is on if Setting is =active= or off if =passive=.

nxt_actions_light_sensor(Port,Value) :-
	not(nxt_actions_input_set(light,Port)), !,
	retractall(nxt_actions_input_set(light,_)),
	nxt_actions_light_sensor_LED(Port,activate),
	assert(nxt_actions_input_set(light,Port)),
	nxt_actions_light_sensor(Port,Value).

nxt_actions_light_sensor(Port,Value) :-
	nxt_actions_input_get(Port,Value), !.


%%	nxt_actions_light_sensor_LED(+Port,+Setting).
%
%	Sets the LED on if Setting is =activate= or off if =passivate=.

nxt_actions_light_sensor_LED(Port,Setting) :-
	not(nxt_actions_input_set(led,Setting)), !,
	retractall(nxt_actions_input_set(led,_)),
	nxt_actions_light_sensor_setting(Setting,Sensor_type),
	nxt_actions_input_setmode(Port,Sensor_type,pctfullscale),
	assert(nxt_actions_input_set(led,Setting)).

nxt_actions_light_sensor_LED(_,_).

nxt_actions_light_sensor_setting(activate,lightactive).
nxt_actions_light_sensor_setting(passivate,lightinactive).


%%	nxt_actions_ultrasonic_sensor(+Port,-Value).
%
%	Gets ultrasonic sensor reading.

nxt_actions_ultrasonic_sensor(Port,Value) :-
	not(nxt_actions_input_set(ultrasonic,Port)), !,
	retractall(nxt_actions_input_set(ultrasonic,_)),
	nxt_actions_input_setmode(Port,lowspeed9V,raw),
	assert(nxt_actions_input_set(ultrasonic,Port)),
	nxt_actions_ultrasonic_sensor(Port,Value).

nxt_actions_ultrasonic_sensor(Port,Value) :-
	% Set single shot.
	nxt_actions_lswrite(Port,0,['\x02','\x41','\x01']),
	% Reading demand.
	nxt_actions_lswrite(Port,1,['\x02','\x42']),
	nxt_actions_lsread(Port,[Value]), !.


%%	nxt_actions_voltage(-Voltage).
%
%	Returns battery voltage in V.

nxt_actions_voltage(_).


%%	nxt_actions_voltage_millivolt(-Voltage).
%
%	Returns battery voltage in mV.

nxt_actions_voltage_millivolt(Voltage) :-
	nxt_actions_connection_type(bt),
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(2),
	nxt_actions_voltage_millivolt_cmd,
	nxt_actions_pop_byte(2),
	nxt_actions_voltage_millivolt_response(Voltage),
	mutex_unlock(nxt_serial_device).

nxt_actions_voltage_millivolt(Voltage) :-
	nxt_actions_connection_type(usb),
	mutex_lock(nxt_serial_device),
	nxt_actions_voltage_millivolt_cmd,
	nxt_actions_voltage_millivolt_response(Voltage),
	mutex_unlock(nxt_serial_device).

nxt_actions_voltage_millivolt_cmd :-
	nxt_actions_send_bytes(['\x00','\x0B']).

nxt_actions_voltage_millivolt_response(Voltage) :-
	nxt_actions_pop_byte(2),
	nxt_actions_check_status_byte, !,
	nxt_actions_get_number(Voltage,2).

nxt_actions_voltage_millivolt_response('error') :-
	nxt_actions_pop_byte(2).


%%	nxt_actions_start_program(+File)
%
%	Starts a Lego executable file on the NXT.

nxt_actions_start_program(_).


%%	nxt_actions_stop_program.
%
%	Stops the currently running Lego executable on the NXT.

nxt_actions_stop_program.


%%	nxt_actions_play_sound_file(+File,+Repeat).
%
%	Plays a sound file from the NXT.
%	Files use the .rso extension.
%	The filename is not case sensitive.
%	Plays sound once if Repeat is true
%	or till nxt_actions_stop_sound_playback is used if Repeat is
%	false.

nxt_actions_play_sound_file(_,true).
nxt_actions_play_sound_file(_,false).


%%	nxt_actions_stop_sound_playback.
%
%	Stops a sound file that has been playing/repeating.

nxt_actions_stop_sound_playback.


%%	nxt_actions_play_tone(Frequency,Duration).
%
%	Plays tone at specified frequency.

nxt_actions_play_tone(Frequency,Duration) :-
	nxt_actions_connection_type(bt),
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(6),
	nxt_actions_play_tone_cmd(Frequency,Duration),
	mutex_unlock(nxt_serial_device).

nxt_actions_play_tone(Frequency,Duration) :-
	nxt_actions_connection_type(usb),
	mutex_lock(nxt_serial_device),
	nxt_actions_play_tone_cmd(Frequency,Duration),
	mutex_unlock(nxt_serial_device).

nxt_actions_play_tone_cmd(Frequency,Duration) :-
	nxt_actions_send_bytes(['\x80','\x03']),
	nxt_actions_send_number(Frequency,2),
	nxt_actions_send_number(Duration,2).


%%	nxt_actions_get_brick_name(-Name).
%
%	Gets the name of the brick.

nxt_actions_get_brick_name(_).


%%	nxt_actions_set_brick_name(+Name)
%
%	Sets the name of the brick.

nxt_actions_set_brick_name(_).





% THE LOWEST LAYER!

% LS COMMUNICATION

%	nxt_actions_lswrite(+Port,+Rx_length,+Bytes).

nxt_actions_lswrite(Port,Rx_length,Bytes) :-
	nxt_actions_connection_type(bt),
	length(Bytes,Tx_length),
	Header_length is Tx_length + 5,
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(Header_length),
	nxt_actions_lswrite_cmd(Port,Tx_length,Rx_length,Bytes),
	mutex_unlock(nxt_serial_device).

nxt_actions_lswrite(Port,Rx_length,Bytes) :-
	nxt_actions_connection_type(usb),
	length(Bytes,Tx_length),
	mutex_lock(nxt_serial_device),
	nxt_actions_lswrite_cmd(Port,Tx_length,Rx_length,Bytes),
	mutex_unlock(nxt_serial_device).

nxt_actions_lswrite_cmd(Port,Tx_length,Rx_length,Bytes) :-
	nxt_actions_input_port(Port,Input_port),
	nxt_actions_send_bytes(['\x80','\x0F',Input_port,Tx_length,Rx_length]),
	nxt_actions_send_bytes(Bytes).


nxt_actions_lsread(Port,Bytes) :-
	nxt_actions_connection_type(bt),
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(3),
	nxt_actions_lsread_cmd(Port),
	nxt_actions_pop_byte(2),
	nxt_actions_lsread_response(Bytes),
	mutex_unlock(nxt_serial_device).

nxt_actions_lsread(Port,Bytes) :-
	nxt_actions_connection_type(usb),
	mutex_lock(nxt_serial_device),
	nxt_actions_lsread_cmd(Port),
	nxt_actions_lsread_response(Bytes),
	mutex_unlock(nxt_serial_device).

nxt_actions_lsread_cmd(Port) :-
	nxt_actions_input_port(Port,Input_port),
	nxt_actions_send_bytes(['\x00','\x10',Input_port]).

nxt_actions_lsread_response(Bytes) :-
	nxt_actions_pop_byte(2),
	nxt_actions_check_status_byte, !,
	nxt_actions_get_byte(Bytes_read),
	nxt_actions_get_bytes(Bytes,Bytes_read),
	Bytes_to_pop is 16 - Bytes_read,
	nxt_actions_pop_byte(Bytes_to_pop).

nxt_actions_lsread_response(['error']) :-
	nxt_actions_pop_byte(17).


% SETOUTPUTSTATE

nxt_actions_output_port('A',0).
nxt_actions_output_port('B',1).
nxt_actions_output_port('C',2).

% Bitfield.
nxt_actions_output_mode(motoron,1).
nxt_actions_output_mode(brake,2).
nxt_actions_output_mode(regulated,4).

nxt_actions_output_regulation(idle,'\x00').
nxt_actions_output_regulation(speed,'\x01').
nxt_actions_output_regulation(sync,'\x02').

% Bitfield?
nxt_actions_output_run(idle,'\x00').
nxt_actions_output_run(rumpup,'\x10').
nxt_actions_output_run(running,'\x20').
nxt_actions_output_run(rumpdown,'\x40').

% Helpers.
nxt_actions_motor_move(Motor,Speed,Mode,Regulation,Turn_ratio,Run,Tacho_limit) :-
	nxt_actions_connection_type(bt),
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(12),
	nxt_actions_motor_move_cmd(Motor,Speed,Mode,Regulation,Turn_ratio,Run,Tacho_limit),
	mutex_unlock(nxt_serial_device).

nxt_actions_motor_move(Motor,Speed,Mode,Regulation,Turn_ratio,Run,Tacho_limit) :-
	nxt_actions_connection_type(usb),
	mutex_lock(nxt_serial_device),
	nxt_actions_motor_move_cmd(Motor,Speed,Mode,Regulation,Turn_ratio,Run,Tacho_limit),
	mutex_unlock(nxt_serial_device).

nxt_actions_motor_move_cmd(Motor,Speed,Mode,Regulation,Turn_ratio,Run,Tacho_limit) :-
	nxt_actions_output_port(Motor,Output_port),
	nxt_actions_send_bytes(['\x80','\x04',Output_port]),
	% Power.
	Power is round(Speed/9),
	nxt_actions_send_sbyte(Power),
	% Mode.
	nxt_actions_motor_move_add_modes(Mode,Mode_byte),
	nxt_actions_send_bytes([Mode_byte]),
	% Regulation mode.
	nxt_actions_output_regulation(Regulation,Regulation_byte),
	nxt_actions_send_bytes([Regulation_byte]),
	% Turn ratio.
	nxt_actions_send_sbyte(Turn_ratio),
	% Run state.
	nxt_actions_output_run(Run,Run_byte),
	nxt_actions_send_bytes([Run_byte]),
	% Tacho limit.
	nxt_actions_send_number(Tacho_limit,4).

nxt_actions_motor_move_add_modes(Mode,Result) :-
	nxt_actions_output_mode(Mode,Result).

nxt_actions_motor_move_add_modes([Mode],Result) :-
	nxt_actions_output_mode(Mode,Result).

nxt_actions_motor_move_add_modes([Mode|Rest],Result) :-
	nxt_actions_output_mode(Mode,Byte),
	nxt_actions_motor_move_add_modes(Rest,Result2),
	Result is Byte + Result2.



%SETINPUTMODE

nxt_actions_input_port('S1',0).
nxt_actions_input_port('S2',1).
nxt_actions_input_port('S3',2).
nxt_actions_input_port('S4',3).

nxt_actions_input_type(nosensor,'\x00').
nxt_actions_input_type(switch,'\x01').
nxt_actions_input_type(temperature,'\x02').
nxt_actions_input_type(reflection,'\x03').
nxt_actions_input_type(angle,'\x04').
nxt_actions_input_type(lightactive,'\x05').
nxt_actions_input_type(lightinactive,'\x06').
nxt_actions_input_type(sounddb,'\x07').
nxt_actions_input_type(sounddba,'\x08').
nxt_actions_input_type(custom,'\x09').
nxt_actions_input_type(lowspeed,'\x0A').
nxt_actions_input_type(lowspeed9V,'\x0B').
nxt_actions_input_type(other,'\x0C').

nxt_actions_input_mode(raw,'\x00').
nxt_actions_input_mode(boolean,'\x20').
nxt_actions_input_mode(transitioncnt,'\x40').
nxt_actions_input_mode(periodcounter,'\x60').
nxt_actions_input_mode(pctfullscale,'\x80').
nxt_actions_input_mode(celsius,'\xA0').
nxt_actions_input_mode(fahrenheit,'\xC0').
nxt_actions_input_mode(anglesteps,'\xE0').
nxt_actions_input_mode(slopemask,'\x1F').
nxt_actions_input_mode(modemask,'\xE0'). %Powtarza sie!

% Helpers
nxt_actions_input_setmode(Port,Type,Mode) :-
	nxt_actions_connection_type(bt),
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(5),
	nxt_actions_input_setmode_cmd(Port,Type,Mode),
	nxt_actions_pop_byte(2),
	nxt_actions_input_setmode_response,
	mutex_unlock(nxt_serial_device).

nxt_actions_input_setmode(Port,Type,Mode) :-
	nxt_actions_connection_type(usb),
	mutex_lock(nxt_serial_device),
	nxt_actions_input_setmode_cmd(Port,Type,Mode),
	nxt_actions_input_setmode_response,
	mutex_unlock(nxt_serial_device).
	
nxt_actions_input_setmode_cmd(Port,Type,Mode) :-
	nxt_actions_input_port(Port,Input_port),
	nxt_actions_send_bytes(['\x00','\x05',Input_port]),
	% Type.
	nxt_actions_input_type(Type,Type_byte),
	nxt_actions_send_bytes([Type_byte]),
	% Mode.
	nxt_actions_input_mode(Mode,Mode_byte),
	nxt_actions_send_bytes([Mode_byte]).
	
nxt_actions_input_setmode_response :-
	nxt_actions_pop_byte(2),
	nxt_actions_check_status_byte, !.

% Error - just unlock mutex and fail.
nxt_actions_input_setmode_response :-
	mutex_unlock(nxt_serial_device), !,
	fail.

% GETINPUTVALUES

% Helpers

nxt_actions_input_get(Port,Value) :-
	nxt_actions_connection_type(bt),
	mutex_lock(nxt_serial_device),
	nxt_actions_send_bt_header(3),
	nxt_actions_input_get_cmd(Port),
	nxt_actions_pop_byte(2),
	nxt_actions_input_get_response(Value),
	mutex_unlock(nxt_serial_device).

nxt_actions_input_get(Port,Value) :-
	nxt_actions_connection_type(usb),
	mutex_lock(nxt_serial_device),
	nxt_actions_input_get_cmd(Port),
	nxt_actions_input_get_response(Value),
	mutex_unlock(nxt_serial_device).

nxt_actions_input_get_cmd(Port) :-
	nxt_actions_input_port(Port,Input_port),
	nxt_actions_send_bytes(['\x00','\x07',Input_port]).

nxt_actions_input_get_response(Value) :-
	nxt_actions_pop_byte(2),
	nxt_actions_check_status_byte, !,
	nxt_actions_pop_byte(9),
	nxt_actions_get_sword(Value),
	nxt_actions_pop_byte(2).

nxt_actions_input_get_response('error') :-
	nxt_actions_pop_byte(13).
	




% HELPERS (sending/rec.../converting bytes).

%	nxt_actions_get_bytes_debug(+Limit).

nxt_actions_get_bytes_debug(1) :- !,
	nxt_actions_get_byte(Byte),
	writeln(Byte).

nxt_actions_get_bytes_debug(N) :-
	nxt_actions_get_byte(Byte),
	writeln(Byte),
	NewN is N-1,
	nxt_actions_get_bytes_debug(NewN).


%	nxt_actions_send_bytes(+Bytes,+Limit).
%
%	Sends list of bytes (Bytes) to the device. Limit says how many bytes has to be sent.
%	If Bytes contains less bytes than it should, rest is filled with zeros.
%	If more, rest is cut.

nxt_actions_send_bytes(_,0) :- !.
nxt_actions_send_bytes([],Limit) :-
	nxt_actions_streams(write,W),
	put_byte(W,'\x00'),
	NewLimit is Limit-1,
	nxt_actions_send_bytes([],NewLimit).
nxt_actions_send_bytes([Byte|Rest],Limit) :-
	nxt_actions_streams(write,W),
	put_byte(W,Byte),
	NewLimit is Limit-1,
	nxt_actions_send_bytes(Rest,NewLimit).


%	nxt_actions_get_bytes(-Bytes,+Limit).
%
%	Gets list of bytes (Bytes) from the device. Limit says how many bytes has to be read.

nxt_actions_get_bytes([Byte],1) :- !,
	nxt_actions_get_byte(Byte).
nxt_actions_get_bytes([Byte|Rest],Limit) :-
	nxt_actions_get_byte(Byte),
	NewLimit is Limit-1,
	nxt_actions_get_bytes(Rest,NewLimit).


%	nxt_actions_send_bytes(+Bytes).
%
%	Sends list of bytes (Bytes) to the device.

nxt_actions_send_bytes([]).
nxt_actions_send_bytes([Byte|Rest]) :-
	nxt_actions_streams(write,W),
	put_byte(W,Byte),
	nxt_actions_send_bytes(Rest).


%	nxt_actions_send_number(+Number,+Limit).
%
%	Sends the Number to the device. Limit says how many bytes is used to send the Number.

nxt_actions_send_number(Number,Limit) :-
	dec_2_bytes(Number,Bytes),
	nxt_actions_send_bytes(Bytes,Limit).


%	nxt_actions_send_sbyte(+Number).
%
%	Sends the Number (range: -128 to 127) to the device.

nxt_actions_send_sbyte(Number) :-
	dec_2_sbyte(Number,Sbyte),
	nxt_actions_send_bytes([Sbyte]).


%	nxt_actions_send_sword(+Number).
%
%	Sends the Number (range: -32768 to 32767) to the device.

nxt_actions_send_sword(Number) :-
	dec_2_sword(Number,Sword),
	nxt_actions_send_number(Sword,2).


%	nxt_actions_send_bt_header(+Length).
%
%	Sends bluetooth header, which has to be attached to every package, when using BT.
%	Length says, how big the package is (without 2 header bytes).

nxt_actions_send_bt_header(Length) :-
	nxt_actions_send_number(Length,2).


%	nxt_actions_pop_byte.
%
%	Gets a single byte from the device and ignores it.

nxt_actions_pop_byte :-
	nxt_actions_pop_byte(1).


%	nxt_actions_pop_byte(+Times).
%
%	Gets bytes (Times says, how many) from the device and ignores them.
	
nxt_actions_pop_byte(1) :- !,
	nxt_actions_streams(read,R),
	get_byte(R,_).
	
nxt_actions_pop_byte(Times) :-
	nxt_actions_streams(read,R),
	get_byte(R,_),
	NewTimes is Times-1,
	nxt_actions_pop_byte(NewTimes).


%	nxt_actions_get_byte(-Byte).
%
%	Reads a single byte from the device.

nxt_actions_get_byte(Byte) :-
	nxt_actions_streams(read,R),
	get_byte(R,Byte).


%	nxt_actions_get_sbyte(-Number).
%
%	Reads a single sbyte from the device.

nxt_actions_get_sbyte(Number) :-
	nxt_actions_get_byte(Byte),
	dec_2_sbyte(Number,Byte).


%	nxt_actions_get_sword(-Number).
%
%	Reads an sword from the device.

nxt_actions_get_sword(Number) :-
	nxt_actions_get_number(Word,2),
	dec_2_sword(Number,Word).


%	nxt_actions_get_slong(-Number).
%
%	Reads an slong from the device.

nxt_actions_get_slong(Number) :-
	nxt_actions_get_number(Long,4),
	dec_2_slong(Number,Long).


%	nxt_actions_check_status_byte.
%
%	Reads a single byte from the device and compares it to 0.
%	Status byte equals 0, when a return package is correct (no errors).

nxt_actions_check_status_byte :-
	nxt_actions_get_byte(Byte),
	Byte = 0.


%	nxt_actions_get_number(-Number,+Limit).
%
%	Reads the Number from the device. Limit says how many bytes is used to send a Number.

nxt_actions_get_number(Number,Limit) :-
	nxt_actions_get_number(Number,Limit,1).

nxt_actions_get_number(Number,B,B) :- !,
	nxt_actions_get_byte(Byte),
	Number is (256**(B-1))*Byte.

nxt_actions_get_number(Number,Limit,B) :-
	nxt_actions_get_byte(Byte),
	NewB is B+1,
	nxt_actions_get_number(NewNumber,Limit,NewB),
	Number is NewNumber + (256**(B-1))*Byte.


%	dec_2_bytes(+Number,-Bytes).
%
%	Converts decimal number to a list of bytes using little endian.

dec_2_bytes(0,[]) :- !.
dec_2_bytes(Number,[Byte|Rest]) :-
	Result is Number//256,
	Byte is Number mod 256,
	dec_2_bytes(Result,Rest).


%	dec_2_sbyte(?Number,?SByte).
%
%	Converts decimal number (range: -128 to 127) to an sbyte (singed byte) and vice versa.

dec_2_sbyte(Number,Sbyte) :-
	var(Sbyte),
	nonvar(Number),
	Number < 0, !,
	Sbyte is Number + 256.

dec_2_sbyte(Number,Sbyte) :-
	var(Number),
	nonvar(Sbyte),
	Sbyte > 127, !,
	Number is Sbyte - 256.

dec_2_sbyte(Number,Number).


%	dec_2_sword(?Number,?Sword).
%
%	Converts decimal number (range: -32768 to 32767) to an sword (singed word) number and vice versa.

dec_2_sword(Number,Sword) :-
	var(Sword),
	nonvar(Number),
	Number < 0, !,
	Sword is Number + 65536.

dec_2_sword(Number,Sword) :-
	var(Number),
	nonvar(Sword),
	Sword > 32767, !,
	Number is Sword - 65536.

dec_2_sword(Number,Number).


%	dec_2_slong(?Number,?Sword).
%
%	Converts decimal number (range: -2147483648 to 2147483647) to an slong (singed long) number and vice versa.

dec_2_slong(Number,Slong) :-
	var(Slong),
	nonvar(Number),
	Number < 0, !,
	Slong is Number + 4294967296.

dec_2_slong(Number,Slong) :-
	var(Number),
	nonvar(Slong),
	Slong > 2147483647, !,
	Number is Slong - 4294967296.

dec_2_slong(Number,Number).
