/** <module> NXT Mindstroms - Actions Sockets.
@authors: 	API: Piotr Hoownia
			Definitions: Marcin Zikowski

Motors: 'A', 'B', 'C'.
Sensors ports: 'S1', 'S2', 'S3', 'S4'.
*/

:- module(nxt_actions_sockets,[
			     nxt_actions_connection_open/0,
			     nxt_actions_connection_open_bt/0,
			     nxt_actions_connection_close/0,
			     nxt_actions_motor_forward/2,
			     nxt_actions_motor_backward/2,
			     nxt_actions_motor_stop/1,
			     nxt_actions_motor_rotate/3,
			     nxt_actions_touch_sensor/2,
			     nxt_actions_sound_sensor/2,
			     nxt_actions_light_sensor/2,
			     nxt_actions_ultrasonic_sensor/2,
			     nxt_actions_start_program/1,
			     nxt_actions_stop_program/0,
			     nxt_actions_play_sound_file/2,
			     nxt_actions_stop_sound_playback/0,
			     nxt_actions_play_tone/2,
			     nxt_actions_get_brick_name/1,
			     nxt_actions_voltage_millivolt/1
			    ]).
			   
:- use_module(library(socket)).
:- use_module(threads).
:- dynamic nxt_actions_sockets/2.

%%	codes_to_string(+Array,-String).
%
%	Changes an array of ASCII codes into a string.

codes_to_string([],'').

codes_to_string([Code|Tail],String):-
	codes_to_string(Tail,String_temp),
	char_code(Char,Code),
	concat(Char,String_temp,String).

%%	nxt_actions_connection_open
%
%	Opens connection.

nxt_actions_connection_open :-
	%win_exec('lib/sockets_server.exe',showdefault),
        tcp_socket(Socket),
        tcp_connect(Socket, '127.0.0.1':2056),
        tcp_open_socket(Socket, ReadFd, WriteFd),
	
	assert(nxt_actions_sockets(ReadFd, WriteFd)),
		
	write_canonical(WriteFd, 'nxt; open'),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _),
		
	write_canonical(WriteFd, 'shortrep'),
	flush_output(WriteFd).
		
%%	nxt_actions_open_connection_bt
% 	
%	Opens BT connection. Only windows OS.
%	DOES NOT WORK PROPERLY!!

nxt_actions_connection_open_bt :-
        tcp_socket(Socket),
        tcp_connect(Socket, '127.0.0.1':2056),
        tcp_open_socket(Socket, ReadFd, WriteFd),
	
	assert(nxt_actions_sockets(ReadFd, WriteFd)),
		
	write_canonical(WriteFd, 'nxt; openbt'),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _),
		
	write_canonical(WriteFd, 'shortrep'),
	flush_output(WriteFd).		

%%	nxt_actions_connection_close
%
%	Closes connection via Bluetooth using iCommand.

nxt_actions_connection_close :-
	nxt_actions_sockets(ReadFd, WriteFd),

	write_canonical(WriteFd, 'nxt; close'),
	flush_output(WriteFd),
		
	write_canonical(WriteFd, 'bye'),
	flush_output(WriteFd),
		
	close(ReadFd),
	close(WriteFd),
		
	retractall(nxt_actions_sockets(_,_)).

%%	nxt_actions_motor_forward(+Motor,+Speed).
%
%	Rotates motor forward at specified speed.

nxt_actions_motor_forward(Motor,Speed) :-
	Power is Speed / 9,
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('motor; setForward;', Motor, A),
	string_concat(A, ';', B),
	string_concat(B, Power, C),
	write_canonical(WriteFd, C),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _).

%%	nxt_actions_motor_forward(+Motor,+Speed).
%
%	Rotates motor backward at specified speed.

nxt_actions_motor_backward(Motor,Speed) :-
	Power is Speed / 9,
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('motor; setReverse;', Motor, A),
	string_concat(A, ';', B),
	string_concat(B, Power, C),
	write_canonical(WriteFd, C),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _).

%%	nxt_motor_stop(+Motor).
%
%	Stops motor.

nxt_actions_motor_stop(Motor) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('motor; stop;', Motor, A),
	string_concat(A, '; false', B),
	write_canonical(WriteFd, B),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _).

%%	nxt_actions_motor_rotate(+Motor,+Power,+Angle).
%
%	Rotates motor at specified speed.
%	Forward if Angle is positive, backward if Angle is negative.
%	Motor will stop, when specified revolution (Angle in degrees) is reached.

nxt_actions_motor_rotate(Motor,Speed,Angle) :-
	nxt_actions_motor_reset(Motor),
	trigger_create(_,nxt_actions_motor_angle_reached(Motor,Angle),nxt_actions_motor_stop(Motor)),
	nxt_actions_motor_rotate_direction(Motor,Speed,Angle).

nxt_actions_motor_rotate_direction(Motor,Speed,Angle) :-
	Angle > 0,
	nxt_actions_motor_forward(Motor,Speed).

nxt_actions_motor_rotate_direction(Motor,Speed,Angle) :-
	Angle < 0,
	nxt_actions_motor_backward(Motor,Speed).

nxt_actions_motor_angle_reached(Motor,Angle) :-
	Angle > 0,
	nxt_actions_motor_get_rotation(Motor,ReachedAngle),
	Angle =< ReachedAngle.

nxt_actions_motor_angle_reached(Motor,Angle) :-
	Angle < 0,
	nxt_actions_motor_get_rotation(Motor,ReachedAngle),
	ReachedAngle =< Angle.

%%	nxt_actions_motor_get_rotation(Motor,Rotation).
%
%	Gets rotation count of the motor in degrees.

nxt_actions_motor_get_rotation(Motor,Rotation) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('motor; getRotationCount;', Motor, A),
	write_canonical(WriteFd, A),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, Rotationtemp),
	codes_to_string(Rotationtemp,RotationAtom),
	atom_number(RotationAtom,Rotation).

%%	nxt_actions_motor_reset(+Motor)
%
%	Resets rotation counter of the Motor.

nxt_actions_motor_reset(Motor) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('motor; resetRotationCount;', Motor, A),
	write_canonical(WriteFd, A),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _).

%%	nxt_actions_touch_sensor(+Port,-Value).
%
%	Gets touch sensor reading.
%	Returns 1 if pressed, 0 otherwise.

nxt_actions_touch_sensor(Port,Value) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('sensor; setTouch;', Port, A),
	write_canonical(WriteFd, A),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _),
		
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('sensor; getValue;', Port, B),
	write_canonical(WriteFd, B),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, Valuetmp),
	codes_to_string(Valuetmp,Value).

%%	nxt_actions_sound_sensor(+Port,-Value).
%
%	Gets sound sensor reading.

nxt_actions_sound_sensor(Port,Value) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('sensor; setSound;', Port, A),
	write_canonical(WriteFd, A),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _),
		
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('sensor; getValue;', Port, B),
	write_canonical(WriteFd, B),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, Valuetmp),
	codes_to_string(Valuetmp,Value).

%%	nxt_actions_light_sensor(+Port,-Value).
%
%	Gets light sensor reading.

nxt_actions_light_sensor(Port,Value) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('sensor; setLight;', Port, A),
	string_concat(A, ';', B),
	string_concat(B, 'false', C),
	write_canonical(WriteFd,C),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _),
		
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('sensor; getValue;', Port, D),
	write_canonical(WriteFd, D),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, Valuetmp),
	codes_to_string(Valuetmp,Value).

%%	nxt_actions_ultrasonic_sensor(+Port,-Value).
%
%	Gets ultrasonic sensor reading.

nxt_actions_ultrasonic_sensor(Port,Value) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('sensor; setSonar;', Port, A),
	write_canonical(WriteFd, A),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _),
	
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('sensor; getSonarValue;', Port, B),
	write_canonical(WriteFd, B),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, Valuetmp),
	codes_to_string(Valuetmp,Value).
		
%%	nxt_actions_start_program(+File)
%
%	Starts a Lego executable file on the NXT.

nxt_actions_start_program(File) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('nxt; startProgram;', File, A),
	write_canonical(WriteFd, A),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _).
		
%%	nxt_actions_stop_program.
%
%	Stops the currently running Lego executable on the NXT.
%	DOES NOT WORK PROPERLY!!

nxt_actions_stop_program :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('nxt; stopProgram', '', A),
	write_canonical(WriteFd, A),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _).
		
%%	nxt_actions_play_sound_file(+File,+Repeat).
%
%	Plays a sound file from the NXT.
%	Files use the .rso extension.
%	The filename is not case sensitive.
%	Plays sound once if Repeat is true
%	or till nxt_actions_stop_sound_playback is used if Repeat is
%	false.

nxt_actions_play_sound_file(File,Repeat) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('nxt; playSoundFile;', File, A),
	string_concat(A,';', B),
	string_concat(B,Repeat, C),
	write_canonical(WriteFd, C),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _).

%%	nxt_actions_stop_sound_playback.
%
%	Stops a sound file that has been playing repeating.

nxt_actions_stop_sound_playback :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('nxt; stopSound', '', A),
	write_canonical(WriteFd, A),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _).
		
%%	nxt_actions_play_tone(Frequency,Duration).
%
%	Plays tone at specified frequency.

nxt_actions_play_tone(Frequency,Duration) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('nxt; playTone;', Frequency, A),
	string_concat(A,';', B),
	string_concat(B,Duration, C),
	write_canonical(WriteFd, C),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, _).
		
%%	nxt_actions_get_brick_name(-Name).
%
%	Gets the name of the brick.
%	DOES NOT WORK PROPERLY!!

nxt_actions_get_brick_name(Name) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	string_concat('nxt; getName', '', A),
	write_canonical(WriteFd, A),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, Valuetmp),
	codes_to_string(Valuetmp,Name).
		
%%	nxt_actions_voltage_millivolt(-Voltage).
%
%	Returns battery voltage in mV.

nxt_actions_voltage_millivolt(Voltage) :-
	nxt_actions_sockets(ReadFd, WriteFd),
	write_canonical(WriteFd, 'nxt; batteryLevel'),
	flush_output(WriteFd),
	read_line_to_codes(ReadFd, Valuetmp),
	codes_to_string(Valuetmp,Voltage).
