%
%     Implementation by Piotr Hoownia <holownia@student.agh.edu.pl>
%
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     PLNXT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of PLNXT.
%
%     PLNXT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     PLNXT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with PLNXT.  If not, see <http://www.gnu.org/licenses/>.
%

:- module(threads,[
			trigger_create/3,
			trigger_create/4,
			trigger_create_noreturn/3,
			trigger_exists/1,
			trigger_kill/1,
			trigger_killall/0,
			timer_create/3,
			timer_create/4,
			timer_exists/1,
			timer_kill/1,
			timer_system_kill/1,
			timer_killall/0,
			timer_system_killall/0,
			wait_till/1
		]).

/** <module> Threads

Implements triggers, timers and simultaneous actions.

@author Piotr Hoownia
@license GNU General Public License
*/


% Cycle interval (time gap between two verifications of an event).
threads_cycle_interval(0.1).
% Fire interval (time gap between fireing an action, if event is still true).
threads_fire_interval(0.2)

:- dynamic threads_trigger/1,
	threads_timer/2.


%%	trigger_create(-ID,+Event,+Action).
%
%	Creates trigger.
%	Action will be fired once, when Event is true.
%	Action can be both predicate and list of predicates.

trigger_create(ID,Event,Action) :-
	trigger_create(ID,Event,Action,1).


%%	trigger_create(-ID,+Event,+Action,+Count).
%
%	Creates trigger.
%	Action will be fired Count times, when Event is true.
%	If Count is =inf= trigger works infinitely.
%	Action can be both predicate and list of predicates.

trigger_create(ID,Event,Action,Count) :-
	thread_create(trigger_start(Event,Action,Count),ID,[detached(true)]).

trigger_start(Event,Action,Count) :-
	thread_self(ID),
	assert(threads_trigger(ID)),
	trigger_cycle(ID,Event,Action,Count).

trigger_cycle(ID,Event,Action,inf) :-
	Event,
	fire(Action),
	threads_fire_interval(Interval),
	sleep(Interval),
	trigger_cycle(ID,Event,Action,inf).

trigger_cycle(ID,Event,Action,1) :-
	Event,
	fire(Action),
	retract(threads_trigger(ID)).

trigger_cycle(ID,Event,Action,Count) :-
	number(Count),
	Count > 1,
	Event,
	NewCount is Count - 1,
	fire(Action),
	threads_fire_interval(Interval),
	sleep(Interval),
	trigger_cycle(ID,Event,Action,NewCount).

trigger_cycle(ID,Event,Action,Count) :-
	threads_trigger(ID),
	threads_cycle_interval(Interval),
	sleep(Interval),
	trigger_cycle(ID,Event,Action,Count).


%%	trigger_create_noreturn(-ID,+Event,+Action).
%
%	Creates trigger.
%	Action will be fired, when Event is true.
%	Trigger will not return to the thread which created it.
%	Action can be both predicate and list of predicates.

trigger_create_noreturn(ID,Event,Action) :-
	thread_self(KillID),
	thread_create(trigger_start_noreturn(KillID,Event,Action),ID,[detached(true)]).
trigger_start_noreturn(KillID,Event,Action) :-
	thread_self(ID),
	assert(threads_trigger(ID)),
	trigger_cycle_noreturn(KillID,ID,Event,Action).

trigger_cycle_noreturn(KillID,ID,Event,Action) :-
	Event,
	with_mutex(nxt_serial_device,thread_signal(KillID,thread_exit('trigger_noreturn'))),
	fire(Action),
	retract(threads_trigger(ID)).

trigger_cycle_noreturn(KillID,ID,Event,Action) :-
	threads_trigger(ID),
	threads_cycle_interval(Interval),
	sleep(Interval),
	trigger_cycle_noreturn(KillID,ID,Event,Action).

%%	trigger_exists(?ID).
%
%	Returns true if trigger exists. Fails if there is no trigger with specified ID.

trigger_exists(ID) :-
	threads_trigger(ID).

%%	trigger_kill(+ID).
%
%	Kills trigger with specified ID.
%	Returns true if trigger has been deleted. Fails if there is no trigger with specified ID.

trigger_kill(ID) :-
	retract(threads_trigger(ID)).

%%	trigger_killall.
%
%	Kills all triggers.

trigger_killall :-
	retractall(threads_trigger(_)).

trigger_killall.

%%	timer_create(-ID,+Time,+Action).
%
%	Creates timer.
%	Action will be fired after specified Time.
%	Action can be both predicate and list of predicates.

timer_create(ID,Time,Action) :-
	thread_create(timer_start(Time,Action,user),ID,[detached(true)]).


%%	timer_create(-ID,+Time,+Action,+Option).
%
%	Creates system timer.
%	Option is:
%
%		* system

timer_create(ID,Time,Action,system) :-
	thread_create(timer_start(Time,Action,system),ID,[detached(true)]).

timer_start(Time,Action,user) :-
	thread_self(ID),
	assert(threads_timer(ID,user)),
	sleep(Time),
	threads_timer(ID,_),
	retract(threads_timer(ID,_)),
	fire(Action).

timer_start(Time,Action,system) :-
	thread_self(ID),
	assert(threads_timer(ID,system)),
	sleep(Time),
	threads_timer(ID,_),
	retract(threads_timer(ID,_)),
	fire(Action).

%%	timer_exists(?ID).
%
%	Returns true if timer exists. Fails if there is no timer with specified ID.

timer_exists(ID) :-
	threads_timer(ID,user).

%%	timer_kill(+ID).
%
%	Kills timer with specified ID.
%	Returns true if timer has been deleted. Fails if there is no timer with specified ID.

timer_kill(ID) :-
	retract(threads_timer(ID,user)).

%%	timer_system_kill(+ID).
%
%	Kills system timer with specified ID.
%	Returns true if timer has been deleted. Fails if there is no timer with specified ID.

timer_system_kill(ID) :-
	retract(threads_timer(ID,system)).

%%	timer_killall.
%
%	Kills all timers.

timer_killall :-
	retractall(threads_timer(_,user)).

timer_killall.

%%	timer_system_killall.
%
%	Kills all system timers.

timer_system_killall :-
	retractall(threads_timer(_,system)).

timer_system_killall.


%%	wait_till(+Event).
%
%	Waits till Event is true.

wait_till(Event) :-
	thread_self(ID),
	trigger_create(_,Event,thread_send_message(ID,resume)),
	%trigger_create(_,Event,[write('Wyslam wiadomosc do '),write(ID),nl,thread_send_message(ID,resume)]),
	%write('Thread '),write(ID),write(' waits for event: '),write(Event),nl,
	thread_get_message(resume).


% Fires action.

fire([Action]) :-
	Action.

fire([Action|Tail]) :-
	Action,
	fire(Tail).

fire(Action) :-
	Action.
